/*
libfawk - A function-only AWK dialect

Copyright (c) 2017..2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libfawk/libfawk.h>
#include <libfawk/execute.h>
#include <libfawk/parser.h>
#include <libfawk/debug.h>
#include <libfawk/array.h>
#include <libfawk/symtab.h>
#include <libfawk/str.h>

#warning TODO: figure how to avoid this
#include <libfawk_fawk/fawk.y.tab.h>

#include <libfawk_fawk/parser_fawk.h>
#include <libfawk_fbas/parser_fbas.h>
#include <libfawk_fpas/parser_fpas.h>

#include "tokens.h"

void libfawk_error(fawk_ctx_t *ctx, const char *str, const char *loc_fn, long loc_line, long loc_col)
{
	fprintf(stderr, "Error: %s at %s %ld:%ld\n", str, loc_fn, loc_line, loc_col);
}

static int getch1(fawk_ctx_t *ctx, fawk_src_t *src)
{
	FILE *f = ctx->parser.isp->user_data;
	return fgetc(f);
}

static int include1(fawk_ctx_t *ctx, fawk_src_t *src, int opening, fawk_src_t *from)
{
	if (opening) {
		FILE *f;
		if ((*src->fn != '/') && (from != NULL)) { /* calculate relative address from 'from' */
			int l1 = strlen(src->fn), l2 = strlen(from->fn);
			char *end, *fn = malloc(l1+l2+4);
			memcpy(fn, from->fn, l2+1);
			end = strrchr(fn, '/');
			if (end != NULL) {
				end++;
				memcpy(end, src->fn, l1+1);
				f = fopen(fn, "r");
			}
			else
				f = fopen(src->fn, "r");
			free(fn);
		}
		else
			f = fopen(src->fn, "r");
		src->user_data = f;
		if (f == NULL) {
			fprintf(stderr, "Can't find %s for include\n", src->fn);
			return -1;
		}
	}
	else
		fclose(src->user_data);

	return 0;
}

FAWK_API int fawk_lex_fawk(fawk_STYPE *lval, fawk_ctx_t *ctx);
static void print_lex(fawk_ctx_t *ctx)
{
	while(!(feof(stdin))) {
		fawk_STYPE lval;
		int tok;
		tok = fawk_lex_fawk(&lval, ctx);
		if (tok >= 0) {
			switch(tok) {
				case T_ID: printf("%s '%s'\n", fawk_tokname[tok], lval.str); break;
				default: printf("%s\n", fawk_tokname[tok]);
			}
		}
		else {
			printf("<EOF>\n");
			break;
		}
	}
}

typedef int (*parser_t)(fawk_ctx_t *ctx);

int main(int argc, char *argv[])
{
	int n, verbose = 0, runlex=0;
	char *funcname = "main";
	char *script = NULL, *exe;
	fawk_ctx_t ctx;
	long runlimit = -1;
	parser_t parser = fawk_parse_fawk;

	/* determine default language */
	exe = strrchr(argv[0], '/');
	if (exe == NULL)
		exe = argv[0];
	else
		exe++;
	if (strncmp(exe, "fawk", 4) == 0) parser = fawk_parse_fawk;
	else if (strncmp(exe, "fbas", 4) == 0) parser = fawk_parse_fbas;
	else if (strncmp(exe, "fpas", 4) == 0) parser = fawk_parse_fpas;

	fawk_init(&ctx);
	ctx.parser.get_char = getch1;
	ctx.parser.include = include1;

	for(n = 1; n < argc; n++) {
		char *cmd = argv[n];
		if (*cmd == '-') {
			while(*cmd == '-') cmd++;
			switch(*cmd) {
				case 'v': verbose++; ctx.exec.trace = 1; break;
				case 'W':
					cmd++;
					if (*cmd == 'f') cmd++; /* for -Wfawk, -Wfbas, -Wfpas */
					switch(*cmd) {
						case 'l': case 'L': runlex = 1; break;
						case 'r': case 'R': runlimit = strtol(argv[n+1], NULL, 10); n++; break;
						case 'a': case 'A': parser = fawk_parse_fawk; break;
						case 'b': case 'B': parser = fawk_parse_fbas; break;
						case 'p': case 'P': parser = fawk_parse_fpas; break;
					}
					break;
			}
		}
		else {
			script = argv[n];
			break; /* leave parsing the rest of the arguments to the argv[] loader */
		}
	}


	/* load the script */
	if (script == NULL) {
		fprintf(stderr, "No script file name specified\n");
		return 1;
	}
	ctx.parser.isp->user_data = fopen(script, "r");
	if (ctx.parser.isp->user_data == NULL) {
		fprintf(stderr, "Can't open fawk script '%s' for read\n", script);
		return 1;
	}
	ctx.parser.isp->fn = fawk_strdup(&ctx, script);

	/* early return if we had to run the lexer only */
	if (runlex) {
		ctx.parser.include = NULL;
		print_lex(&ctx);
		fawk_uninit(&ctx);
		return 0;
	}

	if (parser(&ctx) != 0) {
		fawk_uninit(&ctx);
		return 1;
	}

	if (verbose)
		fawk_dump(&ctx);

	assert(ctx.sp == 0);

	if (fawk_call1(&ctx, funcname) == 0) {
		fawk_cell_t res, *sargv, *v;
		fawk_execret_t er;
		int offs;

		/* fill in ARGV for main() */
		sargv = fawk_push_alloc(&ctx);
		fawk_array_init(&ctx, sargv);
		for(offs = ++n; n < argc; n++) {
			v = fawk_array_resolve_c(&ctx, 1, sargv, FAWK_NUM, (fawk_num_t)(n-offs), FAWK_NIL);
			v->type = FAWK_STR;
			v->data.str = fawk_str_new_from_literal(&ctx, argv[n], -1);
		}

		if (fawk_call2(&ctx, 1) == 0) {
			er = fawk_execute(&ctx, runlimit);
			if (er == FAWK_ER_FIN) {
				fawk_pop(&ctx, &res);
				printf("\nResult: ");
				fawk_dump_cell(&res, 1);
				printf("\n");
				fawk_cell_free(&ctx, &res);
				assert(ctx.sp == 0);
			}
			else
				fprintf(stderr, "Runtime error\n");
		}
		else
			fprintf(stderr, "Failed to call function: main doesn't have ARG as parameter?\n");
	}
	else
		fprintf(stderr, "Failed to call function %s\n", funcname);

	fawk_uninit(&ctx);

	return 0;
}
